package net.jhorstmann.jsonp.buildervscollections;

import java.io.IOException;
import java.io.PushbackReader;
import java.io.StringReader;
import javax.json.JsonArray;
import javax.json.JsonArrayBuilder;
import javax.json.JsonBuilder;
import javax.json.JsonBuilder.JsonBuildable;
import javax.json.JsonObject;
import javax.json.JsonObjectBuilder;
import javax.json.JsonStructure;

/**
 * Trying to get familiar with the proposed API of JSR353,
 * this code uses JsonBuilder inside a recursive descent parser
 * for a simple JSON-like format that consists of arrays,
 * objects with single-character keys and single digit numbers.
 */
public class RecurseiveParsingWithJsonBuilder {

    static void readObject(PushbackReader reader, JsonObjectBuilder objectBuilder) throws IOException {
        int ch = reader.read();
        if (ch == -1) {
            throw new IOException("unexpected eof");
        } else if (ch != '}') {
            reader.unread(ch);
            while (true) {
                ch = reader.read();
                if (ch >= 'a' || ch <= '9') {
                    String key = "" + (char) ch;
                    ch = reader.read();
                    if (ch == -1) {
                        throw new IOException("unexpected eof");
                    } else if (ch != ':') {
                        throw new IOException("unexpected char " + (char) ch);
                    } else {
                        ch = reader.read();
                        if (ch == -1) {
                            throw new IOException("unexpected eof");
                        } else if (ch == '[') {
                            readArray(reader, objectBuilder.startArray(key));
                        } else if (ch == '{') {
                            readObject(reader, objectBuilder.startObject(key));
                        } else if (ch >= '0' && ch <= '9') {
                            objectBuilder.add(key, ch - '0');
                        } else {
                            throw new IOException("unexpected char " + (char) ch);
                        }
                    }
                } else if (ch == -1) {
                    throw new IOException("unexpected eof");
                }

                ch = reader.read();
                if (ch == '}') {
                    break;
                } else if (ch == -1) {
                    throw new IOException("unexpected eof");
                } else if (ch != ',') {
                    throw new IOException("unexpected char " + (char) ch);
                }
            }
        }
    }

    private static void readArray(PushbackReader reader, JsonArrayBuilder arrayBuilder) throws IOException {
        int ch = reader.read();
        if (ch == -1) {
            throw new IOException("unexpected eof");
        } else if (ch != ']') {
            reader.unread(ch);
            while (true) {
                ch = reader.read();
                if (ch == '[') {
                    readArray(reader, arrayBuilder.startArray());
                } else if (ch == '{') {
                    readObject(reader, arrayBuilder.startObject());
                } else if (ch >= '0' && ch <= '9') {
                    arrayBuilder.add(ch - '0');
                } else if (ch == -1) {
                    throw new IOException("unexpected eof");
                } else {
                    throw new IOException("unexpected char " + (char) ch);
                }

                ch = reader.read();
                if (ch == ']') {
                    break;
                } else if (ch == -1) {
                    throw new IOException("unexpected eof");
                } else if (ch != ',') {
                    throw new IOException("unexpected char " + (char) ch);
                }
            }
        }
    }

    /*
     * Parses the root element, either an object or an array.
     */
    public static JsonStructure readRoot(PushbackReader reader) throws IOException {
        int ch = reader.read();
        if (ch == '[') {
            JsonBuilder builder = new JsonBuilder();
            JsonArrayBuilder<JsonBuildable<JsonArray>> arrayBuilder = builder.startArray();
            readArray(reader, arrayBuilder);
            return arrayBuilder.end().build();
        } else if (ch == '{') {
            JsonBuilder builder = new JsonBuilder();
            JsonObjectBuilder<JsonBuildable<JsonObject>> objectBuilder = builder.startObject();
            readObject(reader, objectBuilder);
            return objectBuilder.end().build();
        } else {
            throw new IOException("unexpected char " + (char) ch);
        }
    }

    public static JsonStructure readRoot(String data) throws IOException {
        PushbackReader reader = new PushbackReader(new StringReader(data));
        return readRoot(reader);
    }

    public static void main(String[] args) throws IOException {
        System.out.println(readRoot("[1,2,3,[4],{a:1,b:2,c:{},d:[]},[5,6,{}]]"));
        System.out.println(readRoot("{a:[1,2,3,[4],{a:1,b:2,c:{},d:[]},[5,6,{}]]}"));
    }
}
