package net.jhorstmann.jsonp.xml;

import java.math.BigDecimal;
import javax.json.JsonArray;
import javax.json.JsonArrayBuilder;
import javax.json.JsonBuilder;
import javax.json.JsonBuilder.JsonBuildable;
import javax.json.JsonObject;
import javax.json.JsonObjectBuilder;
import javax.json.JsonStructure;
import net.jhorstmann.jsonp.xmlutil.NestedContentHandler;
import net.jhorstmann.jsonp.xmlutil.TextContentHandler;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

/**
 * ContentHandler implementation that parses XML formatted according to the format described at http://xml.calldei.com/JsonXML into JSON values.
 */
public class JsonHandler extends NestedContentHandler {

    static class MemberHandler extends NestedContentHandler {

        private JsonObjectBuilder builder;
        private String key;

        public MemberHandler(XMLReader reader, NestedContentHandler parent, JsonObjectBuilder builder, String key) {
            super(reader, parent);
            this.builder = builder;
            this.key = key;
        }

        @Override
        public void startNestedElement(String localName, Attributes atts) throws SAXException {
            if ("object".equals(localName)) {
                pushContentHandler(new ObjectHandler(reader, this, builder.startObject(key)));
            } else if ("array".equals(localName)) {
                pushContentHandler(new ArrayHandler(reader, this, builder.startArray(key)));
            } else if ("string".equals(localName)) {
                pushContentHandler(new TextContentHandler(reader, this) {
                    @Override
                    public void endRootElement(String localName) throws SAXException {
                        builder.add(key, getText());
                    }
                });
            } else if ("number".equals(localName)) {
                pushContentHandler(new TextContentHandler(reader, this) {
                    @Override
                    public void endRootElement(String localName) throws SAXException {
                        builder.add(key, new BigDecimal(getText()));
                    }
                });
            } else if ("null".equals(localName)) {
                pushContentHandler(new TextContentHandler(reader, this) {
                    @Override
                    public void endRootElement(String localName) throws SAXException {
                        builder.addNull(key);
                    }
                });
            } else if ("boolean".equals(localName)) {
                pushContentHandler(new TextContentHandler(reader, this) {
                    @Override
                    public void endRootElement(String localName) throws SAXException {
                        builder.add(key, Boolean.parseBoolean(getText()));
                    }
                });
            }
        }
    }

    static class ObjectHandler extends NestedContentHandler {

        private JsonObjectBuilder builder;

        public ObjectHandler(XMLReader reader, NestedContentHandler parent, JsonObjectBuilder builder) {
            super(reader, parent);
            this.builder = builder;
        }

        @Override
        public void startNestedElement(String localName, Attributes atts) throws SAXException {
            if ("member".equals(localName)) {
                String key = atts.getValue("name");
                pushContentHandler(new MemberHandler(reader, this, builder, key));
            }
        }
    }

    static class ArrayHandler extends NestedContentHandler {

        private JsonArrayBuilder builder;

        public ArrayHandler(XMLReader reader, NestedContentHandler parent, JsonArrayBuilder builder) {
            super(reader, parent);
            this.builder = builder;
        }

        @Override
        public void startNestedElement(String localName, Attributes atts) throws SAXException {
            if ("object".equals(localName)) {
                pushContentHandler(new ObjectHandler(reader, this, builder.startObject()));
            } else if ("array".equals(localName)) {
                pushContentHandler(new ArrayHandler(reader, this, builder.startArray()));
            } else if ("string".equals(localName)) {
                pushContentHandler(new TextContentHandler(reader, this) {
                    @Override
                    public void endRootElement(String localName) throws SAXException {
                        builder.add(getText());
                    }
                });
            } else if ("number".equals(localName)) {
                pushContentHandler(new TextContentHandler(reader, this) {
                    @Override
                    public void endRootElement(String localName) throws SAXException {
                        System.out.println("text " + getText());
                        builder.add(new BigDecimal(getText()));
                    }
                });
            } else if ("null".equals(localName)) {
                pushContentHandler(new TextContentHandler(reader, this) {
                    @Override
                    public void endRootElement(String localName) throws SAXException {
                        builder.addNull();
                    }
                });
            } else if ("boolean".equals(localName)) {
                pushContentHandler(new TextContentHandler(reader, this) {
                    @Override
                    public void endRootElement(String localName) throws SAXException {
                        builder.add(Boolean.parseBoolean(getText()));
                    }
                });
            }
        }
    }
    private JsonStructure structure;
    private JsonObjectBuilder<JsonBuildable<JsonObject>> objectBuilder;
    private JsonArrayBuilder<JsonBuildable<JsonArray>> arrayBuilder;

    public JsonHandler(XMLReader reader) {
        super(reader);
    }

    public JsonStructure getStructure() {
        if (structure == null) {
            if (objectBuilder != null) {
                structure = objectBuilder.end().build();
            } else if (arrayBuilder != null) {
                structure = arrayBuilder.end().build();
            } else {
                throw new IllegalStateException();
            }
        }
        return structure;
    }

    @Override
    public void startNestedElement(String localName, Attributes atts) throws SAXException {
        System.out.println("JsonHandler.startElement");
        structure = null;
        JsonBuilder builder = new JsonBuilder();
        if ("object".equals(localName)) {
            objectBuilder = builder.startObject();
            pushContentHandler(new ObjectHandler(reader, this, objectBuilder));
        } else if ("array".equals(localName)) {
            arrayBuilder = builder.startArray();
            pushContentHandler(new ArrayHandler(reader, this, arrayBuilder));
        } else {
            throw new SAXException("Unknown root element " + localName);
        }
    }

    @Override
    public void endDocument() throws SAXException {
        System.out.println("JsonHandler.endDocument");
    }

    @Override
    public void endRootElement(String localName) throws SAXException {
        System.out.println("JsonHandler.endRootElement");
        /*
        if ("object".equals(localName)) {
            structure = objectBuilder.endObject().build();
        } else if ("array".equals(localName)) {
            structure = arrayBuilder.endArray().build();
        }
        */
    }
}
