package net.jhorstmann.jsonp.xml;

import net.jhorstmann.jsonp.xmlutil.AbstractXMLReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.math.BigDecimal;
import java.util.LinkedList;
import javax.json.spi.JsonProvider;
import javax.json.stream.JsonParser;
import javax.json.stream.JsonParser.Event;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;

/**
 * XMLReader implementation that reports JSON data according to the format defined at http://xml.calldei.com/JsonXML
 */
public class JsonXMLReader extends AbstractXMLReader {

    static enum State {

        OBJECT {
            @Override
            void startStructure(ContentHandler handler) throws SAXException {
                handler.startElement("", "object", "object", EMPTY_ATTRIBUTES);
            }

            @Override
            void endStructure(ContentHandler handler) throws SAXException {
                handler.endElement("", "object", "object");
            }

            @Override
            void startItem(ContentHandler handler, String key) throws SAXException {
                AttributesImpl attr = new AttributesImpl();
                attr.addAttribute("", "name", "name", "CDATA", key);
                handler.startElement("", "member", "member", attr);
            }

            @Override
            void endItem(ContentHandler handler) throws SAXException {
                handler.endElement("", "member", "member");
            }
        },
        ARRAY {
            @Override
            void startStructure(ContentHandler handler) throws SAXException {
                handler.startElement("", "array", "array", EMPTY_ATTRIBUTES);
            }

            @Override
            void endStructure(ContentHandler handler) throws SAXException {
                handler.endElement("", "array", "array");
            }
        };

        abstract void startStructure(ContentHandler handler) throws SAXException;

        abstract void endStructure(ContentHandler handler) throws SAXException;

        void startItem(ContentHandler handler, String key) throws SAXException {
        }

        void endItem(ContentHandler handler) throws SAXException {
        }

        private void characters(ContentHandler handler, String string) throws SAXException {
            if (string != null) {
                char[] chars = string.toCharArray();
                handler.characters(chars, 0, chars.length);
            }
        }

        private static String elementNameForValue(Event event) {
            switch (event) {
                case VALUE_NUMBER:
                    return "number";
                case VALUE_STRING:
                    return "string";
                case VALUE_TRUE:
                case VALUE_FALSE:
                    return "boolean";
                case VALUE_NULL:
                    return "null";
                default:
                    throw new IllegalArgumentException("Invalid value type " + event);
            }
        }

        void value(ContentHandler handler, JsonParser.Event event, String content) throws SAXException {
            String elem = elementNameForValue(event);
            handler.startElement("", elem, elem, EMPTY_ATTRIBUTES);
            characters(handler, content);
            handler.endElement("", elem, elem);
        }

        void addString(ContentHandler handler, String string) throws SAXException {
            value(handler, Event.VALUE_STRING, string);
        }

        void addNumber(ContentHandler handler, BigDecimal number) throws SAXException {
            value(handler, Event.VALUE_NUMBER, number.toPlainString());
        }

        void addBoolean(ContentHandler handler, boolean bool) throws SAXException {
            if (bool) {
                value(handler, Event.VALUE_TRUE, "true");
            } else {
                value(handler, Event.VALUE_FALSE, "false");
            }
        }

        void addNull(ContentHandler handler) throws SAXException {
            value(handler, Event.VALUE_NULL, null);
        }
    }

    @Override
    public void parse(Reader reader) throws IOException, SAXException {
        JsonParser parser = JsonProvider.provider().createParser(reader);
        parse(parser);
    }

    @Override
    public void parse(InputStream in) throws IOException, SAXException {
        JsonParser parser = JsonProvider.provider().createParser(in);
        parse(parser);
    }

    void parse(JsonParser parser) throws SAXException {
        LinkedList<State> stack = new LinkedList<State>();
        State state = null;
        String currentKey = null;
        if (parser.hasNext()) {
            contentHandler.startDocument();
            while (parser.hasNext()) {
                Event event = parser.next();
                System.out.println(state + " " + event + " " + stack);
                switch (event) {
                    case KEY_NAME:
                        currentKey = parser.getString();
                        break;
                    case START_OBJECT:
                        if (state != null) {
                            state.startItem(contentHandler, currentKey);
                        }
                        state = State.OBJECT;
                        stack.addFirst(state);
                        state.startStructure(contentHandler);
                        break;
                    case START_ARRAY:
                        if (state != null) {
                            state.startItem(contentHandler, currentKey);
                        }
                        state = State.ARRAY;
                        stack.addFirst(state);
                        state.startStructure(contentHandler);
                        break;
                    case END_OBJECT:
                        state = stack.removeFirst();
                        assert (state == State.OBJECT);
                        state.endStructure(contentHandler);
                        if (!stack.isEmpty() && (state = stack.getFirst()) != null) {
                            state.endItem(contentHandler);
                        }
                        break;
                    case END_ARRAY:
                        state = stack.removeFirst();
                        assert (state == State.ARRAY);
                        state.endStructure(contentHandler);
                        if (!stack.isEmpty() && (state = stack.getFirst()) != null) {
                            state.endItem(contentHandler);
                        }
                        break;
                    case VALUE_STRING:
                        assert (state != null);
                        assert (state == State.ARRAY || currentKey != null);
                        state.startItem(contentHandler, currentKey);
                        state.addString(contentHandler, parser.getString());
                        state.endItem(contentHandler);
                        break;
                    case VALUE_NUMBER:
                        assert (state != null);
                        assert (state == State.ARRAY || currentKey != null);
                        state.startItem(contentHandler, currentKey);
                        state.addNumber(contentHandler, parser.getBigDecimalValue());
                        state.endItem(contentHandler);
                        break;
                    case VALUE_TRUE:
                    case VALUE_FALSE:
                        assert (state != null);
                        assert (state == State.ARRAY || currentKey != null);
                        state.startItem(contentHandler, currentKey);
                        state.addBoolean(contentHandler, event == Event.VALUE_TRUE);
                        state.endItem(contentHandler);
                        break;
                    case VALUE_NULL:
                        assert (state != null);
                        assert (state == State.ARRAY || currentKey != null);
                        state.startItem(contentHandler, currentKey);
                        state.addNull(contentHandler);
                        state.endItem(contentHandler);
                        break;
                }
            }
            contentHandler.endDocument();
        }
    }
}